# Control Page

{{
control_writer.write_control_with_sections(control, profile, group_title,
['statement', 'objective', 'ExpectedEvidence', 'ImplGuidance', 'table_of_parameters'],
{'statement':'Statement Header',
'objective':'Control Objective Header',
'ExpectedEvidence':'Control Expected Evidence Header',
'ImplGuidance':'Implementation Guidance',
'table_of_parameters':'Table of Control Parameters'
},
label_column=True,
add_group_to_title=True
)
| safe
}}
