# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Validate Command."""
import argparse
import logging

from trestle import __version__
from trestle.core.commands.command_docs import CommandBase
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.oscal import OSCAL_VERSION

logger = logging.getLogger(__name__)


class VersionCmd(CommandBase):
    """Output version info for trestle and OSCAL."""

    name = 'version'

    def _run(self, args: argparse.Namespace) -> int:
        version_string = f'Trestle version v{__version__} based on OSCAL version {OSCAL_VERSION}'
        self.out(version_string)
        return CmdReturnCodes.SUCCESS.value
