cd $TRESTLE_BASEDIR/catalogs/mycatalog

# Merges the properties inside each of the files passed in via the -f option to a destination file specified with the -d option.
trestle merge -e catalog.*


# This command results in the contents of original step-merged_catalog